from scrapy import Request


class SeleniumRequest(Request):

    def __init__(self,
                 driver_func: callable = None,
                 driver_func_args: tuple = (),
                 driver_func_kwargs: dict = {},
                 driver=None,
                 keep_driver=False, *args, **kwargs):
        """
        SeleniumRequest constructor

        You can send some function, that will be running after page loading.
        The function must have one required parameter of the selenium.webdriver.remote.webdriver.WebDriver type.
        Also you can send optional args and kwargs to the function

        Args:
            driver_func(callable): function, that will be running after page loading
            driver_func_args(tuple): Tuple of args
            driver_func_kwargs(dict): Dict of kwargs
            driver(http.Driver): use kept driver (see keep_driver)
            keep_driver(bool): don't unblock driver after using. Driver will be keeped as meta['driver']. Use with caution
            *args:
            **kwargs:
        """
        self.is_selenium = True
        self.driver_func = driver_func
        self.driver_func_args = driver_func_args
        self.driver_func_kwargs = driver_func_kwargs
        self.driver = driver
        self.keep_driver = keep_driver
        if driver is not None:
            kwargs['url'] = driver.web_driver.current_url
            kwargs['dont_filter'] = True
        super().__init__(*args, **kwargs)



import os
import re
from zipfile import ZipFile
import shutil


def create_extension(username, password):
    path = os.path.dirname(os.path.realpath(__file__))
    with open(f'{path}/extension/background.js', 'w') as file:
        text = """
                function callbackFn(details) {
                    return {
                        authCredentials: {
                            username: "%s",
                            password: "%s"
                        }
                    };
                }
                
                browser.webRequest.onAuthRequired.addListener(
                        callbackFn,
                        {urls: ["<all_urls>"]},
                        ['blocking']
                );
               """ % (username, password)
        file.write(text)
    zip_file = ZipFile(f'{path}/extensions/extension.xpi', 'w')
    zip_file.write(f'{path}/extension/background.js', arcname='background.js')
    zip_file.write(f'{path}/extension/manifest.json', arcname='manifest.json')
    zip_file.close()


def create_anticaptcha_extension(api_key):
    path = os.path.dirname(os.path.realpath(__file__))
    with ZipFile(f'{path}/anticaptcha_plugin/anticaptcha-plugin_v0.56.xpi', 'r') as zip_file:
        zip_file.extractall(f'{path}/anticaptcha_plugin/anticaptcha-plugin')
    with open(f'{path}/anticaptcha_plugin/anticaptcha-plugin/js/config_ac_api_key.js') as file:
        js_text = ''.join(file.readlines())
    js_text = re.sub("antiCapthaPredefinedApiKey = ''", f"antiCapthaPredefinedApiKey = '{api_key}'", js_text)
    with open(f'{path}/anticaptcha_plugin/anticaptcha-plugin/js/config_ac_api_key.js', 'w') as file:
        file.write(js_text)
    shutil.make_archive(f'{path}/extensions/anticaptcha-plugin.xpi', 'zip',
                        f'{path}/anticaptcha_plugin/anticaptcha-plugin')
    os.rename(f'{path}/extensions/anticaptcha-plugin.xpi.zip',
              f'{path}/extensions/anticaptcha-plugin.xpi')



import os
import re
from itertools import cycle

# import seleniumwire.webdriver
from selenium.webdriver.remote.command import Command
from selenium import webdriver
from bs4 import BeautifulSoup
from webdriver_manager.firefox import GeckoDriverManager
from .extension_creator import create_extension, create_anticaptcha_extension


class ChangeProxyMixin:
    """Mixin with methods for scraping on selenium.webdriver(Firefox) base"""

    def __init__(self,
                 change_proxies_on_each_request=True,
                 proxies=None,
                 install_adblock=True,
                 anticaptcha_api_key=None,
                 *args, **kwargs):
        self.path = os.path.dirname(os.path.realpath(__file__))
        self.change_proxies_on_each_request = change_proxies_on_each_request
        self.proxies = proxies
        if self.proxies:
            self.proxies = cycle(self.proxies)
        kwargs['executable_path'] = GeckoDriverManager().install()
        super().__init__(*args, **kwargs)
        self.execute(Command.GET, {'url': "about:config"})
        # need for install extensions
        self.set_preference('xpinstall.signatures.required', 'false')
        if install_adblock:
            self.install_addon(
                f'{self.path}/extensions/adblocker_ultimate-3.7.10-an+fx.xpi')
        if anticaptcha_api_key:
            create_anticaptcha_extension(anticaptcha_api_key)
            self.install_addon(
                f'{self.path}/extensions/anticaptcha-plugin.xpi')

    def soup(self):
        """Get soup from page"""
        return BeautifulSoup(self.page_source, 'lxml')

    def change_proxy(self, proxy):
        """Open config page and change proxy"""
        proxy = re.split(':|@', proxy)
        proxy_username = None
        proxy_password = None
        proxy_type = proxy[0]
        proxy_address = proxy[-2]
        proxy_port = int(proxy[-1])
        if len(proxy) > 3:
            proxy_username = proxy[1].strip('//')
            proxy_password = proxy[2]
        self.execute(Command.GET, {'url': "about:config"})
        if 'socks' in proxy_type:
            self.set_preference('network.proxy.socks_version', int(proxy_type[-1]))
            self.set_preference('network.proxy.socks', proxy_address)
            self.set_preference('network.proxy.socks_port', proxy_port)
            self.set_preference('network.proxy.type', 1)
        elif 'https' in proxy[0].lower():
            self.set_preference('network.proxy.ssl', proxy_address)
            self.set_preference('network.proxy.ssl_port', proxy_port)
            self.set_preference('network.proxy.type', 1)
        elif 'http' in proxy[0].lower():
            self.set_preference('network.proxy.http', proxy_address)
            self.set_preference('network.proxy.ftp', proxy_address)
            self.set_preference('network.proxy.socks', proxy_address)
            self.set_preference('network.proxy.ssl', proxy_address)
            self.set_preference('network.proxy.http_port', proxy_port)
            self.set_preference('network.proxy.ftp_port', proxy_port)
            self.set_preference('network.proxy.socks_port', proxy_port)
            self.set_preference('network.proxy.ssl_port', proxy_port)
            self.set_preference('network.proxy.type', 1)
            self.set_preference('network.proxy.share_proxy_settings', 'true')
        if proxy_username and proxy_password:
            create_extension(proxy_username, proxy_password)
            self.install_addon(
                f'{self.path}/extensions/extension.xpi')

    def disable_cache(self):
        """Disable browser cache"""
        self.execute(Command.GET, {'url': "about:config"})
        self.set_preference('browser.cache.disk.enable', 'false')
        self.set_preference('browser.cache.memory.enable', 'false')
        self.set_preference('browser.cache.offline.enable', 'false')
        self.set_preference('network.http.use-cache', 'false')

    def set_preference(self, pref, params):
        """Set preference in 'about:config' """
        if params in ['false', 'true']:
            self.execute_script('Components.classes["@mozilla.org/preferences-service;1"]\
                .getService(Components.interfaces.nsIPrefBranch).setBoolPref("' + pref + '", ' + str(params) + ');')
        elif type(params) == int:
            self.execute_script('Components.classes["@mozilla.org/preferences-service;1"]\
                .getService(Components.interfaces.nsIPrefBranch).setIntPref("' + pref + '", ' + str(params) + ');')
        elif type(params) == str:
            self.execute_script('Components.classes["@mozilla.org/preferences-service;1"]\
                .getService(Components.interfaces.nsIPrefBranch).setCharPref("' + pref + '", "' + params + '");')

    def get(self, url):
        """Loads a web page in the current browser session."""
        if self.change_proxies_on_each_request and self.proxies and url != 'about:config':
            proxy = self.proxies.__next__()
            self.change_proxy(proxy)
        self.execute(Command.GET, {'url': url})


class ScrapyWebdriver(ChangeProxyMixin, webdriver.Firefox):
    pass


# class ScrapyWebdriverWire(ChangeProxyMixin, seleniumwire.webdriver.Firefox):
#     pass


if __name__ == '__main__':
    from proxies import proxies
    driver = ScrapyWebdriver(proxies=proxies)




import logging

from ..scrapy_webdriver import ScrapyWebdriver

from scrapy.utils.python import to_bytes
from scrapy.http import HtmlResponse

from twisted.internet.threads import deferToThreadPool
from twisted.internet import reactor
from twisted.internet.defer import Deferred

from .http import SeleniumRequest


logger = logging.getLogger(__name__)


class Driver:
    def __init__(self,
                 pool: 'DriverPool' = None,
                 change_proxy_on_each_request: bool = True,
                 proxies: tuple = (),
                 install_adblock: bool = True,
                 anticaptcha_api_key: str = None):
        self.web_driver = ScrapyWebdriver(change_proxies_on_each_request=change_proxy_on_each_request,
                                          proxies=proxies,
                                          install_adblock=install_adblock,
                                          anticaptcha_api_key=anticaptcha_api_key)
        self._blocked = False
        self.pool = pool

    @property
    def blocked(self):
        return self._blocked

    def block(self):
        self._blocked = True

    def unblock(self):
        self._blocked = False
        self.pool.update(self)


class DriverPool:
    def __init__(self, size=1,
                 change_proxy_on_each_request=True,
                 proxies=(),
                 install_adblock=True,
                 anticaptcha_api_key=None):
        self.size = size
        self.drivers = []
        self._waiting = []
        self.chang_proxy_on_each_request = change_proxy_on_each_request
        self.proxies = proxies
        self.install_adblock = install_adblock

    def append_driver(self):
        driver = Driver(pool=self,
                        change_proxy_on_each_request=self.chang_proxy_on_each_request,
                        proxies=self.proxies,
                        install_adblock=self.install_adblock)
        self.drivers.append(driver)
        return driver

    def close(self):
        for driver in self.drivers:
            driver.web_driver.close()

    def update(self, driver):
        """ Listener callback for DriverItem"""
        if self._waiting:
            free_driver = self._waiting.pop()
            free_driver.callback(driver)
            logger.info('Free deferred from waiting, len = %s', len(self._waiting))

    def get_driver(self, keeped_driver=None):
        if keeped_driver is not None:
            free_driver = Deferred()
            free_driver.callback(keeped_driver)
            return free_driver
        free_driver = Deferred()
        free_drivers = [driver for driver in self.drivers if not driver.blocked]
        if free_drivers:
            free_driver.callback(free_drivers[0])
            logger.info("Free driver found")
        elif len(self.drivers) < self.size:
            driver = self.append_driver()
            logger.info("Free driver not found. New driver appended")
            free_driver.callback(driver)
        else:
            logger.info("Free driver not found. Waiting for ....")
            self._waiting.append(free_driver)
            logger.info('Append deferred tor waiting, len = %s', len(self._waiting))
        return free_driver

    @staticmethod
    def _get_url(driver: Driver, request: SeleniumRequest) -> HtmlResponse:
        web_driver = driver.web_driver
        if request.driver is None:
            web_driver.get(request.url)
        if request.driver_func is not None:
            func_res = request.driver_func(web_driver,
                                           *request.driver_func_args,
                                           **request.driver_func_kwargs)
            if func_res:
                request.meta['driver_func_res'] = func_res
        body = to_bytes(web_driver.page_source)  # body must be of type bytes
        response = HtmlResponse(web_driver.current_url, body=body, encoding='utf-8', request=request)
        response.meta['driver'] = driver
        return response

    @staticmethod
    def _response(driver: Driver, request: SeleniumRequest) -> Deferred:
        driver.block()
        deferred = deferToThreadPool(reactor, reactor.getThreadPool(), DriverPool._get_url, driver, request)
        return deferred

    @staticmethod
    def _unblock(response: HtmlResponse):
        driver = response.meta['driver']
        if not response.request.keep_driver:
            driver.unblock()
            response.meta['driver'] = None
        return response

    def get_response(self, request: SeleniumRequest) -> Deferred:
        driver = self.get_driver(request.driver)
        response = driver.addCallback(self._response, request).addCallback(self._unblock)
        return response



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="scrapy_webdriver",
	version="0.56",
	author="Aleksey Kvyatkovsky",
	author_email="kvyatkovsky@mail.ru",
	description="Class based on selenium webdriver.Firefox with methods  for scraping.",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/KvyatkovskyAleksey/ScrapeWebdriver",
	packages=setuptools.find_packages(),
	install_requires=[
		'beautifulsoup4==4.9.3',
		'selenium==3.141.0',
		'webdriver-manager==3.4.2',
		'Scrapy==2.5.0',
		# 'selenium_wire==4.3.1'
    ],
	classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],)


from scrapy.crawler import Crawler
from scrapy import signals

from .middleware_utils.http import SeleniumRequest
from .middleware_utils.drivers import DriverPool


class AsyncSeleniumMiddleware:
    def __init__(self,
                 drivers_number: int,
                 proxies: tuple,
                 change_proxy_on_each_request: bool,
                 install_adblock: bool):
        self.driver_pool = DriverPool(drivers_number,
                                      proxies=proxies,
                                      change_proxy_on_each_request=change_proxy_on_each_request,
                                      install_adblock=install_adblock)

    @classmethod
    def from_crawler(cls, crawler: Crawler):
        drivers_number = crawler.settings.getint("SELENIUM_POOL_SIZE", 1)
        proxies = crawler.settings.get("SELENIUM_PROXY_POOL", ())
        change_proxy_on_each_request = crawler.settings.get("SELENIUM_CHANGE_PROXY_ON_EACH_REQUEST", True)
        install_adblock = crawler.settings.get("SELENIUM_INSTALL_ADBLOCK", True)
        anticaptcha_api_key = crawler.settings.get("SELENIUM_ANTICAPTCHA_API_KEY", None)
        middleware = cls(drivers_number, proxies, change_proxy_on_each_request, install_adblock)
        crawler.signals.connect(middleware.spider_closed, signals.spider_closed)
        return middleware

    def process_request(self, request: SeleniumRequest, spider):
        try:
            request.is_selenium
        except AttributeError:
            return None

        return self.driver_pool.get_response(request)

    def spider_closed(self, spider):
        self.driver_pool.close()
